@@ -6,7 +6,7 @@
    Foundation.
 */
 
-static char Id[] = "$Id: pnp-select.c,v 0.1 1998/10/09 21:23:00 fox Exp $";
+static char Id[] = "$Id: pnp-select.c,v 0.2 1998/11/10 20:52:59 fox Exp $";
 
 #include <stdio.h>
 #include <malloc.h>
@@ -84,7 +84,7 @@ alloc_resources ( struct resource *res, int count,
 	        return 1;
 	    }
 	} else {
-	    if (allocate_resource(res->type, res->value, res->size)) {
+	    if (allocate_resource(res->type, res->value, res->size, "pnpdump")) {
 	        if (alloc_resources(res+1,count-1, parent_res, parent_count))
 		    return 1;
 		deallocate_resource(res->type, res->value, res->size);
