@@ -36,7 +36,7 @@
 #define TMP_LEN 16
 #define LARGE_LEN 65536
 
-static char rcsid[] = "$Id: pnpdump.c,v 1.16 1998/10/09 22:19:06 fox Exp $";
+static char rcsid[] = "$Id: pnpdump.c,v 1.18 1999/02/14 22:47:18 fox Exp $";
 
 #ifdef REALTIME
 /*
@@ -231,6 +231,8 @@ prepare_to_print(void)
 }
 
 unsigned char serial_identifier[NUM_CARDS + 1][IDENT_LEN];
+unsigned char csumdiff[NUM_CARDS + 1];
+
 unsigned char tmp[TMP_LEN];
 unsigned char large[LARGE_LEN];
 static char *devstring = 0;
@@ -253,6 +255,7 @@ void read_idents(void);
 static void usage(char *program_name);
 
 int do_dumpregs = 0;
+int do_ignorecsum = 0;
 void dumpregs(int);
 
 char *devidstr(unsigned char, unsigned char, unsigned char, unsigned char);
@@ -277,11 +280,12 @@ main(int argc, char **argv)
 		{"script", optional_argument, NULL, 's'},
 		{"reset", no_argument, NULL, 'r'},
 		{"dumpregs", no_argument, NULL, 'd'},
+		{"ignorecsum", no_argument, NULL, 'i'},
 		{0, 0, 0, 0},
 	};
 	int opt;
 
-	while ((opt = getopt_long(argc, argv, "cms:rd", longopts, NULL)) != EOF)
+	while ((opt = getopt_long(argc, argv, "cms:rdi", longopts, NULL)) != EOF)
 	{
 		switch (opt)
 		{
@@ -301,6 +305,9 @@ main(int argc, char **argv)
 		case 'd':
 			do_dumpregs = 1;
 			break;
+		case 'i':
+			do_ignorecsum = 1;
+			break;
 		case '?':
 #if 0
 			fprintf(stderr, "unrecognized option.\n");
@@ -342,6 +349,10 @@ main(int argc, char **argv)
 		{
 			do_autoconfig = 1;
 		}
+		else if (strcmp(argv[1], "--ignorecsum") == 0)
+		{
+			do_ignorecsum = 1;
+		}
 		else
 		{
 			break;
@@ -349,7 +360,7 @@ main(int argc, char **argv)
 		argv++;
 		argc--;
 	}
-#endif /* __DJGPP */
+#endif /* __DJGPP__ */
 
 	if ((argc < 1) || (argc > 3))
 	{
@@ -400,7 +411,7 @@ main(int argc, char **argv)
 		   "#\n"
 		   "# For details of this file format, see isapnp.conf(5)\n"
 		   "#\n"
-		   "# For latest information on isapnp and pnpdump see:\n"
+		   "# For latest information and FAQ on isapnp and pnpdump see:\n"
 		   "# http://www.roestock.demon.co.uk/isapnptools/\n"
 		   "#\n"
 		   "# Compiler flags:", rcsid);
@@ -416,12 +427,14 @@ main(int argc, char **argv)
 #ifdef NEEDNANOSLEEP
 	printf(" -DNEEDNANOSLEEP");
 #endif /* NEEDNANOSLEEP */
+#ifdef ABORT_ONRESERR
+	printf(" -DABORT_ONRESERR");
+#endif /* ABORT_ONRESERR */
 	printf("\n#\n");
+	/* Have to do this anyway to check readport for clashes */
+	alloc_system_resources();
+	/* Now get board info */
 	initiate();
-	if (do_autoconfig)
-	{
-		alloc_system_resources();
-	}
 	resources = NULL;
 	resource_count = 0;
 	for (i = 1; i <= boards_found; i++)
@@ -504,13 +517,18 @@ usage(char *program_name)
 			"http://www.roestock.demon.co.uk/isapnptools/\n"
 			"This software has NO WARRANTY, use at your OWN RISK\n\n"
 #ifndef __DJGPP__
-			"Usage: %s [--masks] [--config] [--reset] [--script=outputfile] [--dumpregs]\n",
+			"Usage: %s [--masks] [--config] [--reset] [--script=outputfile] [--dumpregs] [--ignorecsum]\n",
 #else
-			"Usage: %s [--masks] [--config] [--reset] [--dumpregs]\n",
+			"Usage: %s [--masks] [--config] [--reset] [--dumpregs] [--ignorecsum]\n",
 #endif
 			rcsid, program_name);
 }
 
+/*
+ * Wait for resource data byte
+ *
+ * Return true if error occurred (timeout)
+ */
 int
 statuswait(void)
 {
@@ -570,8 +588,9 @@ sendkey(void)
 	ADDRESS(0);
 	ADDRESS(0);
 	for (i = 0; i < INIT_LENGTH; i++)
+	{
 		ADDRESS(initdata[i]);
-	delaynus(2000L);
+	}
 }
 
 void
@@ -583,20 +602,27 @@ initiate(void)
 		/* All cards now isolated, read the first one */
 		for (port = read_port; port <= MAX_READ_ADDR; port += READ_ADDR_STEP)
 		{
+			/* Check it doesn't clash with anything */
+			if(!allocate_resource(IOport_TAG, port, 1, "readport"))
+			{
+				continue;
+			}
+			deallocate_resource(IOport_TAG, port, 1);
 			/* Make sure all cards are in Wait for key */
 			CONFIGCONTROL;
 			WRITE_DATA(CONFIG_WAIT_FOR_KEY);
 			/* Make them listen */
+			delaynus(2000L);
 			sendkey();
 			/* Reset the cards */
 			CONFIGCONTROL;
+			WRITE_DATA(CONFIG_RESET_CSN);
 			if(do_fullreset)
-				WRITE_DATA(CONFIG_RESET_DRV);
-			else
-				WRITE_DATA(CONFIG_RESET_CSN | CONFIG_WAIT_FOR_KEY);
-			delaynus(2000L);
-			/* Send the key again */
-			sendkey();
+			{
+				CONFIGCONTROL;
+				WRITE_DATA(CONFIG_RESET);
+				delaynus(2000L);
+			}
 			printf("# Trying port address %04x\n", port);
 			if (do_isolate())
 				break;
@@ -617,10 +643,14 @@ initiate(void)
 				printf(" %02x", serial_identifier[boards_found][i]);
 			printf("\n");
 		}
+		printf("\n# (DEBUG)\n(READPORT 0x%04x)\n", port);
+		if(do_ignorecsum)
+			printf("(IGNORECRC)\n");
 		if(do_fullreset)
-			printf("\n# (DEBUG)\n(READPORT 0x%04x)\n(ISOLATE CLEAR)\n(IDENTIFY *)\n", port);
+			printf("(ISOLATE CLEAR)\n");
 		else
-			printf("\n# (DEBUG)\n(READPORT 0x%04x)\n(ISOLATE PRESERVE)\n(IDENTIFY *)\n", port);
+			printf("(ISOLATE PRESERVE)\n");
+		printf("(IDENTIFY *)\n");
 	}
 	else
 	{
@@ -676,7 +706,7 @@ do_isolate(void)
 	SetRdPort(port);
 	delaynus(1000L);
 	SERIALISOLATION;
-	delaynus(1000L);
+	delaynus(2000L);
 	for (index = 0; index < IDENT_LEN - 1; index++)
 	{
 		for (i = 0; i < 8; i++)
@@ -695,10 +725,10 @@ do_isolate(void)
 				}
 				else
 				{
-					goodaddress = 0;
+					goodaddress = 0; 
 				}
 			}
-			/* printf("%02x %02x - ",c1,c2); */
+			/* printf("\n %02x %02x - bit %02x, goodaddress %d",c1,c2, newbit, goodaddress); */
 			serial_identifier[boards_found][index] >>= 1;
 			serial_identifier[boards_found][index] |= newbit;
 			/* Update checksum */
@@ -707,8 +737,7 @@ do_isolate(void)
 			csum >>= 1;
 			csum |= newbit;
 		}
-		/* printf("%02x ", serial_identifier[boards_found][index]); */
-		/* printf("\n"); */
+		/* printf(" *** %02x \n", serial_identifier[boards_found][index]); */
 	}
 	/* printf("computed csum is %02x", csum); */
 	for (i = 0; i < 8; i++)
@@ -726,23 +755,19 @@ do_isolate(void)
 				newbit = 0x80;
 			}
 		}
-		/* printf("%02x %02x | ",c1,c2); */
+		/* printf("\n %02x %02x - bit %02x, goodaddress %d",c1,c2, newbit, goodaddress); */
 		serial_identifier[boards_found][index] >>= 1;
 		serial_identifier[boards_found][index] |= newbit;
 	}
-	/* printf("\n"); */
-	if (goodaddress && (csum == serial_identifier[boards_found][index]))
+	/* printf(" *** csum *** %02x\n", serial_identifier[boards_found][index]); */
+	if (goodaddress && (do_ignorecsum || (csum == serial_identifier[boards_found][index])))
 	{
 		CARDSELECTNUMBER;
 		WRITE_DATA(boards_found);
+		if((csumdiff[boards_found] = (csum - serial_identifier[boards_found][index])))
+			printf("# WARNING: serial identifier mismatch for board %d: expected 0x%02X, got 0x%02X\n", boards_found, serial_identifier[boards_found][index], csum);
 		return (1);
 	}
-#ifdef NOTACHANCE
-	else
-	{
-		printf("csum = %02x\n", csum);
-	}
-#endif
 	/* We didn't find one */
 	boards_found--;
 	return (0);
@@ -779,7 +804,7 @@ lengtherror(int len, char *msg)
 #ifdef ABORT_ONRESERR
 #define BREAKORCONTINUE   goto oncorruptresourcedata
 #else
-#define BREAKORCONTINUE   continue
+#define BREAKORCONTINUE   break
 #endif
 
 static void
@@ -826,13 +851,14 @@ unread_resource_data(char prev)
 static inline int
 read_resource_data(unsigned char *result)
 {
-	if (statuswait())
-		return 1;
 	if (backspaced)
 	{
 		*result = backspaced_char;
 		backspaced = 0;
+		return 0;
 	}
+	if (statuswait())
+		return 1;
 	*result = RESOURCEDATA;
 	csum = (csum + *result) & 0xFF;
 	return 0;
@@ -1044,6 +1070,8 @@ read_board_resources(struct resource *first_element,
 	*result_array_ptr = result_array;
 }
 
+static int IORangeCheck = 0;
+
 static void
 print_resource(FILE * output_file, struct resource *res, int selected)
 {
@@ -1084,6 +1112,7 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 			}
 			if (res->len != IDENT_LEN)
 			{
+				i = 0;
 				printf("# Ident byte %d, (%02x) differs from resource data (%02x)\n", i, serial_identifier[csn][i], res->data[0]);
 				printf("#Assuming the card is broken and this is the start of the resource data\n");
 			}
@@ -1132,6 +1161,14 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 			indep = "";
 			if (nestdepth)
 			{
+				/* If we have a device name, show it (the last description string before the END DF flag is it) */
+				if (serialno == 0xffffffffUL)
+					printf(" (NAME \"%s/-1[%d]", devid, logdevno-1);
+				else
+					printf(" (NAME \"%s/%lu[%d]", devid, serialno, logdevno-1);
+				if(devstring)
+					printf("{%-20s}", devstring);
+				printf("\")\n");
 				printf("%s (ACT Y)\n", comment_if_not_selected);
 				while (nestdepth)
 				{
@@ -1142,15 +1179,22 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 			}
 			printf("#\n# %sLogical device id %s\n", indep, devidstr(res->data[0], res->data[1], res->data[2], res->data[3]));
 			indep = "    ";
-			for (i = 1, reg = 2; reg < 256; i++, reg <<= 1)
+			IORangeCheck = 0;
+			for (i = 0, reg = 1; reg < 256; i++, reg <<= 1)
 			{
 				if (res->data[4] & reg)
+				{
 					printf("# %s%s\n", indep, regfns[i]);
+					if(i == 1)
+						IORangeCheck = 1;
+				}
 			}
 			for (i = 0, reg = 1; reg < 256; i++, reg <<= 1)
 			{
 				if (res->data[5] & reg)
+				{
 					printf("# %sDevice supports vendor reserved register @ 0x%02x\n", indep, 0x38 + i);
+				}
 			}
 			printf("#\n# Edit the entries below to uncomment out the configuration required.\n");
 			printf("# Note that only the first value of any range is given, this may be changed if required\n");
@@ -1333,14 +1377,6 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 								  selected && (res->value == i));
 			}
 			indep = "    ";
-			/* If we have a device name, show it (the last description string before the END DF flag is it) */
-			if (serialno == 0xffffffffUL)
-				printf(" (NAME \"%s/-1[%d]", devid, logdevno-1);
-			else
-				printf(" (NAME \"%s/%lu[%d]", devid, serialno, logdevno-1);
-			if(devstring)
-				printf("{%-20s}", devstring);
-			printf("\")\n");
 			printf("\n# %sEnd dependent functions\n", indep);
 			break;
 		}
@@ -1363,8 +1399,8 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 				printf("# %s%s0x%04x..0x%04x\n", indep, indep, indep, i, i + res->data[6] - 1);
 			}
 #endif /* DUMPADDR */
-			printf("%s (IO %d (SIZE %d) (BASE 0x%04lx))\n",
-				   comment_if_not_selected, curio, res->data[6], res->value);
+			printf("%s (IO %d (SIZE %d) (BASE 0x%04lx)%s)\n",
+				   comment_if_not_selected, curio, res->data[6], res->value, IORangeCheck ? " (CHECK)" : "");
 			curio++;
 			if (!starteddeps)
 				depio = curio;
@@ -1380,8 +1416,8 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 			}
 			printf("# %sFixed IO base address 0x%04lx\n", indep, res->start);
 			printf("# %s%sNumber of IO addresses required: %ld\n", indep, indep, res->size);
-			printf("%s (IO %d (SIZE %ld) (BASE 0x%04lx))\n",
-				   comment_if_not_selected, curio, res->size, res->start);
+			printf("%s (IO %d (SIZE %ld) (BASE 0x%04lx)%s)\n",
+				   comment_if_not_selected, curio, res->size, res->start, IORangeCheck ? " (CHECK)" : "");
 			curio++;
 			if (!starteddeps)
 				depio = curio;
@@ -1452,9 +1488,9 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 				exit(1);
 			}
 			for (i = 0; i < len; i++)
-				putchar(devstring[i] = res->data[i]);
+				devstring[i] = res->data[i];
 			devstring[i] = 0;
-			printf("<--\n");
+			printf("%s<--\n", devstring);
 			break;
 		}
 	case UNICODEstr_TAG:
@@ -1490,6 +1526,14 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 			char *indep = "";
 			if (nestdepth)
 			{
+				/* If we have a device name, show it (the last description string before the END DF flag is it) */
+				if (serialno == 0xffffffffUL)
+					printf(" (NAME \"%s/-1[%d]", devid, logdevno-1);
+				else
+					printf(" (NAME \"%s/%lu[%d]", devid, serialno, logdevno-1);
+				if(devstring)
+					printf("{%-20s}", devstring);
+				printf("\")\n");
 				printf("%s (ACT Y)\n", comment_if_not_selected);
 				while (nestdepth)
 				{
@@ -1787,8 +1831,8 @@ read_board(int csn, struct resource **res_list, int *list_len)
 	/* Get the first byte */
 	if (read_resource_data(&res_array[array_len - 1].data[0]) != 0)
 		return;
-	/* Just check the first byte, if these match we assume it's ok */
-	if (res_array[array_len - 1].data[0] != serial_identifier[csn][0])
+	/* Just check the first byte, if these match, or the serial identifier had a checksum error, we assume it's ok */
+	if((!csumdiff[csn])&&(res_array[array_len - 1].data[0] != serial_identifier[csn][0]))
 	{
 		/*
 		 * Assume the card is broken and this is the start of the resource
@@ -1808,6 +1852,40 @@ read_board(int csn, struct resource **res_list, int *list_len)
 		if (read_resource_data(&res_array[array_len - 1].data[i]) != 0)
 			return;
 	}
+	if(csumdiff[csn])
+	{
+		/* Original serial identifier had checksum error, so make the serial
+		 * identifier what we read from the resource data, as that's what
+		 * isapnp will see from an (identify *), assuming checksum ok
+		 */
+		int data;
+		int newbit;
+		int j;
+		csum = 0x6a;
+		for (i = 0; i < IDENT_LEN - 1; i++)
+		{
+			data = res_array[array_len - 1].data[i];
+			for(j = 0; j < 8; j++)
+			{
+				newbit = (data & 1) << 7;
+				data >>= 1;
+				if(((csum >> 1) ^ csum) & 1)
+					newbit ^= 0x80;
+				csum >>= 1;
+				csum |= newbit;
+			}
+		}
+		if(csum == res_array[array_len - 1].data[IDENT_LEN - 1])
+		{
+			printf("# WARNING: Making Board %d serial identifier the resource data version\n", csn);
+			for (i = 0; i < IDENT_LEN; i++)
+			{
+				serial_identifier[csn][i] = res_array[array_len - 1].data[i];
+			}
+		}
+		else
+			printf("# WARNING: Board %d resource data identifier has checksum error too\n", csn);
+	}
 	/* Now for the actual resource data */
 	csum = 0;
   broken:
